# flake8:noqa
from ledgereth._meta import author, email, version
from ledgereth.accounts import find_account, get_account_by_path, get_accounts
from ledgereth.messages import sign_message, sign_typed_data_draft
from ledgereth.objects import (
    SignedTransaction,
    SignedType1Transaction,
    SignedType2Transaction,
    Transaction,
    Type1Transaction,
    Type2Transaction,
)
from ledgereth.transactions import create_transaction, sign_transaction

__all__ = [
    "Transaction",
    "Type1Transaction" "Type2Transaction" "SignedTransaction",
    "SignedType1Transaction",
    "SignedType2Transaction",
    "create_transaction",
    "find_account",
    "get_account_by_path",
    "get_accounts",
    "sign_message",
    "sign_transaction",
    "sign_typed_data_draft",
]
__version__ = version
__author__ = author
__email__ = email



import binascii
import struct
from typing import Optional, Union

from ledgereth.comms import Dongle, dongle_send_data, init_dongle
from ledgereth.constants import DATA_CHUNK_SIZE, DEFAULT_CHAIN_ID, DEFAULT_PATH_STRING
from ledgereth.objects import SignedMessage, SignedTypedMessage
from ledgereth.utils import (
    chunks,
    coerce_access_list,
    is_bip32_path,
    is_hex_string,
    parse_bip32_path,
)

AnyText = Union[str, bytes]


def sign_message(
    message: AnyText,
    sender_path: str = DEFAULT_PATH_STRING,
    dongle: Optional[Dongle] = None,
) -> SignedMessage:
    """Sign a simple text message.  Message will be prefixed by the Ethereum
    app on the Ledger device according to `EIP-191`_.

    :param message: (:code:`str|bytes`) - A bit of text to sign
    :param sender_path: (:code:`str`) - HID derivation path for the account to
        sign with.
    :param dongle: (:class:`ledgerblue.Dongle.Dongle`) - The Web3 instance to use
    :return: :class:`ledgereth.objects.SignedMessage`

    .. _`EIP-191`: https://eips.ethereum.org/EIPS/eip-191
    """
    given_dongle = dongle is not None
    dongle = init_dongle(dongle)
    retval = None

    if type(message) == str:
        message = message.encode("utf-8")

    # Silence mypy due to type cohersion above
    assert type(message) == bytes

    encoded = struct.pack(">I", len(message))
    encoded += message

    if not is_bip32_path(sender_path):
        raise ValueError("Invalid sender BIP32 path given to sign_transaction")

    path = parse_bip32_path(sender_path)
    payload = (len(path) // 4).to_bytes(1, "big") + path + encoded

    chunk_count = 0
    for chunk in chunks(payload, DATA_CHUNK_SIZE):
        chunk_size = len(chunk)

        if chunk_count == 0:
            retval = dongle_send_data(
                dongle,
                "SIGN_MESSAGE_FIRST_DATA",
                chunk,
                Lc=chunk_size.to_bytes(1, "big"),
            )
        else:
            retval = dongle_send_data(
                dongle,
                "SIGN_MESSAGE_SECONDARY_DATA",
                chunk,
                Lc=chunk_size.to_bytes(1, "big"),
            )
        chunk_count += 1

    if retval is None or len(retval) < 64:
        raise Exception("Invalid response from Ledger")

    v = int(retval[0])
    r = int(binascii.hexlify(retval[1:33]), 16)
    s = int(binascii.hexlify(retval[33:65]), 16)

    signed = SignedMessage(message, v, r, s)

    # If this func inited the dongle, then close it, otherwise core dump
    if not given_dongle:
        dongle.close()

    return signed


def sign_typed_data_draft(
    domain_hash: AnyText,
    message_hash: AnyText,
    sender_path: str = DEFAULT_PATH_STRING,
    dongle: Optional[Dongle] = None,
) -> SignedTypedMessage:
    """Sign `EIP-721`_ typed data.

    .. DANGER::
        EIP-712 is still in DRAFT status and APIs may change, including the
        Ledger app-ethereum implementation.

    :param domain_hash: (:code:`str`) - Hash of the EIP-712 domain
    :param message_hash: (:code:`str`) - Hash of the message
    :param sender_path: (:code:`str`) - HID derivation path for the account to
        sign with. Defaults to first account in the derivation path.
    :param dongle: (:class:`ledgerblue.Dongle.Dongle`) -  The Dongle instance to
        use to communicate with the Ledger device
    :return: :class:`ledgereth.objects.SignedTypedMessage` Signed message object

    For a real example of usage, see how this is used with `eth_account`_ in
    `ledgereth's unit tests`_.

    .. _`EIP-721`: https://eips.ethereum.org/EIPS/eip-712
    .. _`eth_account`: https://eth-account.readthedocs.io/
    .. _`ledgereth's unit tests`: https://github.com/mikeshultz/ledger-eth-lib/blob/2e47e7b9d70136a6dda0229c7bf516ed6bbe850f/tests/test_message_signing.py#L55-L74
    """

    given_dongle = dongle is not None
    dongle = init_dongle(dongle)
    retval = None

    if type(domain_hash) == str:
        domain_hash = domain_hash.encode("utf-8")
    if type(message_hash) == str:
        message_hash = message_hash.encode("utf-8")

    # Silence mypy due to type cohersion above
    assert type(domain_hash) == bytes
    assert type(message_hash) == bytes

    encoded = domain_hash + message_hash

    if not is_bip32_path(sender_path):
        raise ValueError("Invalid sender BIP32 path given to sign_transaction")

    path = parse_bip32_path(sender_path)
    payload = (len(path) // 4).to_bytes(1, "big") + path + encoded

    retval = dongle_send_data(
        dongle,
        "SIGN_TYPED_DATA",
        payload,
        Lc=len(payload).to_bytes(1, "big"),
    )

    if retval is None or len(retval) < 64:
        raise Exception("Invalid response from Ledger")

    v = int(retval[0])
    r = int(binascii.hexlify(retval[1:33]), 16)
    s = int(binascii.hexlify(retval[33:65]), 16)

    signed = SignedTypedMessage(domain_hash, message_hash, v, r, s)

    # If this func inited the dongle, then close it, otherwise core dump
    if not given_dongle:
        dongle.close()

    return signed



import argparse
import sys
from enum import IntEnum

from eth_utils import decode_hex, encode_hex

from ledgereth import (
    create_transaction,
    find_account,
    get_account_by_path,
    get_accounts,
    sign_message,
    sign_typed_data_draft,
)
from ledgereth.comms import init_dongle


class ExitCodes(IntEnum):
    SUCCESS = 0
    GENERAL_ERROR = 1
    INVALID_ARGUMENT = 2


def get_args(argv):
    parser = argparse.ArgumentParser(description="Do some ledger ops")
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Print extra debugging information",
    )

    subparsers = parser.add_subparsers(
        title="Commands", dest="command", help="Available commands"
    )

    # `accounts` command
    accounts_parser = subparsers.add_parser(
        "accounts", help="Print accounts from the Ledger"
    )
    accounts_parser.add_argument(
        "path",
        metavar="PATH",
        nargs="?",
        help="Get the account for a specific path",
    )
    accounts_parser.add_argument(
        "-c",
        "--count",
        type=int,
        default=3,
        help="How many accounts to fetch (default: 3)",
    )

    # `send` command
    send_parser = subparsers.add_parser(
        "send", help="Send a value transaction from a Ledger account"
    )
    send_parser.add_argument(
        "from_address",
        metavar="FROM",
        help="Account to sign with",
    )
    send_parser.add_argument(
        "to_address",
        metavar="TO",
        help="Account to send to",
    )
    send_parser.add_argument(
        "wei",
        metavar="WEI",
        type=int,
        help="Amount to send (in wei)",
    )
    send_parser.add_argument(
        "-n",
        "--nonce",
        type=int,
        required=True,
        help="Nonce to use for the transaction",
    )
    send_parser.add_argument(
        "-c",
        "--chainid",
        type=int,
        default=1,
        help="Chain ID (default: 1)",
    )
    send_parser.add_argument(
        "-g",
        "--gas",
        type=int,
        default=22000,
        help="The gas limit to use for the tx (default: 22000)",
    )
    send_parser.add_argument(
        "-p",
        "--gasprice",
        type=int,
        help="The gas price to use for the tx",
    )
    send_parser.add_argument(
        "-f",
        "--max-fee",
        type=int,
        help="The max fee per gas to use for the tx",
    )
    send_parser.add_argument(
        "-b",
        "--priority-fee",
        type=int,
        default=0,
        help="The priority fee per gas to use for the tx (default: 0)",
    )
    send_parser.add_argument(
        "-d",
        "--data",
        type=str,
        help="The hex data to send with the tx (default: empty)",
    )

    # `sign` command
    sign_parser = subparsers.add_parser(
        "sign", help="Sign a text message with a Ledger account using EIP-191 v0"
    )
    sign_parser.add_argument(
        "account_address",
        metavar="ADDRESS",
        help="Address of the account to sign with",
    )
    sign_parser.add_argument(
        "message",
        metavar="MESSAGE",
        help="Message to sign",
    )

    # `signtyped` command
    signtyped_parser = subparsers.add_parser(
        "signtyped", help="Sign a text message with a Ledger account using EIP-191 v0"
    )
    signtyped_parser.add_argument(
        "account_address",
        metavar="ADDRESS",
        help="Address of the account to sign with",
    )
    signtyped_parser.add_argument(
        "domain_hash",
        metavar="DOMAIN",
        help="Domain hash to sign",
    )
    signtyped_parser.add_argument(
        "message_hash",
        metavar="MESSAGE",
        help="Message hash to sign",
    )

    return parser.parse_args(argv)


def print_accounts(dongle, args):
    if args.path:
        account = get_account_by_path(args.path, dongle)
        print(f"Account {account.path} {account.address}")
    else:
        accounts = get_accounts(dongle, count=args.count)
        for i, a in enumerate(accounts):
            print(f"Account {i}: {a.path} {a.address}")


def send_value(dongle, args):
    print(f"Sending {args.wei} ETH from {args.from_address} to {args.to_address}")

    account = find_account(args.from_address, dongle)

    if not account:
        print("Account not found on device", file=sys.stderr)
        dongle.close()
        sys.exit(ExitCodes.INVALID_ARGUMENT)

    if not args.gasprice and not args.max_fee:
        print("Either --gasprice or --max-fee must be provided", file=sys.stderr)
        dongle.close()
        sys.exit(ExitCodes.INVALID_ARGUMENT)

    to_address = args.to_address

    signed = create_transaction(
        destination=to_address,
        amount=args.wei,
        gas=args.gas,
        gas_price=args.gasprice,
        max_fee_per_gas=args.max_fee,
        max_priority_fee_per_gas=args.priority_fee,
        data=args.data or "",
        nonce=args.nonce,
        chain_id=args.chainid,
        sender_path=account.path,
    )

    print(f"Signed Raw Transaction: {signed.raw_transaction()}")


def sign_text_message(dongle, args):
    print(f'Signing "{args.message}" with {args.account_address}')

    account = find_account(args.account_address, dongle)
    signed = sign_message(args.message, account.path)

    print(f"Signature: {signed.signature}")


def sign_typed_data(dongle, args):
    print(f"Signing typed data with account {args.account_address}")
    print(f"Domain hash: {args.domain_hash}")
    print(f"Message hash: {args.message_hash}")

    account = find_account(args.account_address, dongle)
    signed = sign_typed_data_draft(
        decode_hex(args.domain_hash), decode_hex(args.message_hash), account.path
    )

    print(f"Signature: {signed.signature}")


COMMANDS = {
    "accounts": print_accounts,
    "send": send_value,
    "sign": sign_text_message,
    "signtyped": sign_typed_data,
}


def main(argv=sys.argv[1:]):
    args = get_args(argv)
    command = args.command

    if command not in COMMANDS:
        print(f"Invalid command: {command}", file=sys.stderr)
        sys.exit(ExitCodes.INVALID_ARGUMENT)

    dongle = init_dongle(debug=args.debug)
    COMMANDS[command](dongle, args)
    dongle.close()


if __name__ == "__main__":
    main()



# Some of the following imports utilize web3.py deps that are not deps of
# ledgereth.
from eth_account.messages import encode_structured_data
from eth_utils import decode_hex, encode_hex
from rlp import encode

from ledgereth.accounts import find_account, get_accounts
from ledgereth.messages import sign_message, sign_typed_data_draft
from ledgereth.objects import SignedTransaction
from ledgereth.transactions import create_transaction
from ledgereth.utils import decode_web3_access_list

"""
ACCOUNT DERIVATION ISSUES

Derivation path debate is ever ongoing.  Ledger Live app uses opton #3 here.
The old chrome app used #4.

1) 44'/60'/0'/0/x
2) 44'/60'/0'/x/0
3) 44'/60'/x'/0/0
4) 44'/60'/0'/x

The Ledger Live account enumeration algo appears to be:

1) Try legacy account X (if no balance, GOTO 3)
2) X+1 and GOTO 1
3) Try new-derivation Y (if no balance, RETURN)
4) Y+1 and GOTO 3

Since this library is trying not to resort to JSON-RPC calls, this algorithm is
not usable, so it's either or, and it currently defaults to the newer
derivation.

To use legacy derivation, set the environment variable LEDGER_LEGACY_ACCOUNTS

Ref (cannot find an authoritative source):
https://github.com/ethereum/EIPs/issues/84#issuecomment-292324521
"""


class LedgerSignerMiddleware:
    """Web3.py middleware.  It will automatically intercept the relevant
    JSON-RPC calls and respond with data from your Ledger device.

    :Intercepted JSON-RPC methods:

    - ``eth_sendTransaction`` (``web3.eth.send_transaction()``)
    - ``eth_accounts`` (``web3.eth.accounts``)
    - ``eth_sign`` (``web3.eth.sign()``)
    - ``eth_signTypedData``  (``web3.eth.sign_typed_data()``)

    :Example:

    .. code:: python

        >>> from web3.auto import w3
        >>> from ledgereth.web3 import LedgerSignerMiddleware
        >>> w3.middleware_onion.add(LedgerSignerMiddleware)
        >>> w3.eth.accounts
        ['0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266', '0x8C8d35429F74ec245F8Ef2f4Fd1e551cFF97d650', '0x98e503f35D0a019cB0a251aD243a4cCFCF371F46']

    """

    _dongle = None

    def __init__(self, make_request, w3):
        self.w3 = w3
        self.make_request = make_request

    def __call__(self, method, params):
        if method == "eth_sendTransaction":
            return self._handle_eth_sendTransaction(method, params)

        elif method == "eth_accounts":
            return self._handle_eth_accounts(method, params)

        elif method == "eth_sign":
            return self._handle_eth_sign(method, params)

        elif method == "eth_signTypedData":
            return self._handle_eth_signTypedData(method, params)

        # Send on to the next middleware(s)
        return self.make_request(method, params)

    def _handle_eth_accounts(self, method, params):
        """Handler for eth_accounts RPC calls"""
        return {
            "result": list(map(lambda a: a.address, get_accounts(dongle=self._dongle))),
        }

    def _handle_eth_sendTransaction(self, method, params):
        """Handler for eth_sendTransaction RPC calls"""
        new_params = []

        for tx_obj in params:
            sender_address = tx_obj.get("from")
            nonce = tx_obj.get("nonce")
            gas = tx_obj.get("gas")
            gas_price = tx_obj.get("gasPrice")
            max_fee_per_gas = tx_obj.get("maxFeePerGas")
            max_priority_fee_per_gas = tx_obj.get("maxPriorityFeePerGas")
            value = tx_obj.get("value", "0x00")
            access_list = None

            if not sender_address:
                # TODO: Should this use a default?
                raise ValueError('"from" field not provided')

            if not gas:
                # TODO: What's the default web3.py behavior for this?
                raise ValueError('"gas" field not provided')

            if not gas_price and not max_fee_per_gas:
                raise ValueError('"gasPrice" or "maxFeePerGas" field not provided')

            sender_account = find_account(sender_address, dongle=self._dongle)

            if not sender_account:
                raise Exception(f"Account {sender_address} not found")

            if nonce is None:
                nonce = self.w3.eth.get_transaction_count(sender_address)

            if "accessList" in tx_obj:
                access_list = decode_web3_access_list(tx_obj["accessList"])

            signed_tx = create_transaction(
                chain_id=self.w3.eth.chain_id,
                destination=tx_obj.get("to"),
                amount=int(value, 16),
                gas=int(gas, 16),
                gas_price=int(gas_price, 16) if gas_price else None,
                max_fee_per_gas=int(max_fee_per_gas, 16) if max_fee_per_gas else None,
                max_priority_fee_per_gas=int(max_priority_fee_per_gas, 16)
                if max_priority_fee_per_gas
                else None,
                nonce=nonce,
                data=tx_obj.get("data", b""),
                sender_path=sender_account.path,
                access_list=access_list,
                dongle=self._dongle,
            )

            new_params.append(signed_tx.rawTransaction)

        # Change to raw tx call
        method = "eth_sendRawTransaction"
        params = new_params

        return self.make_request(method, params)

    def _handle_eth_sign(self, mehtod, params):
        """Handler for eth_sign RPC calls"""
        if len(params) != 2:
            raise ValueError("Unexpected RPC request params length for eth_sign")

        account = params[0]
        message = decode_hex(params[1])

        signer_account = find_account(account, dongle=self._dongle)
        signed = sign_message(message, signer_account.path, dongle=self._dongle)

        return {
            "result": signed.signature,
        }

    def _handle_eth_signTypedData(self, mehtod, params):
        """Handler for eth_signTypedData RPC calls"""
        if len(params) != 2:
            raise ValueError("Unexpected RPC request params length for eth_sign")

        account = params[0]
        typed_data = params[1]

        if type(typed_data) != dict:
            raise TypeError(
                "Expected type data to be a dictionary for second param for eth_signTypedData call"
            )

        # Use eth_account to encode and hash the typed data
        signable = encode_structured_data(typed_data)
        domain_hash = signable.header
        message_hash = signable.body

        # Find the account and sign with Ledger
        signer_account = find_account(account, dongle=self._dongle)
        signed = sign_typed_data_draft(
            domain_hash, message_hash, signer_account.path, dongle=self._dongle
        )

        return {
            "result": signed.signature,
        }



from typing import Optional

from eth_utils import add_0x_prefix
from ledgerblue.comm import getDongle
from ledgerblue.commException import CommException
from ledgerblue.Dongle import Dongle

from ledgereth.constants import DEFAULT_PATH_ENCODED
from ledgereth.exceptions import LedgerError
from ledgereth.objects import ISO7816Command

DONGLE_CACHE: Optional[Dongle] = None
DONGLE_CONFIG_CACHE: Optional[bytes] = None


class LedgerCommands:
    """APDU commands for communication with ledger-app-eth.  Tested only on a Ledger Nano S.

    See `ledger-app-eth<https://github.com/LedgerHQ/ledger-app-eth/blob/master/doc/ethapp.asc>`_
    documentation.
    """

    GET_CONFIGURATION = ISO7816Command(
        CLA=b"\xe0",
        INS=b"\x06",
        P1=b"\x00",
        P2=b"\x00",
        Lc=b"\x00",
        Le=b"\x04",
    )

    GET_DEFAULT_ADDRESS_NO_CONFIRM = ISO7816Command(
        CLA=b"\xe0",
        INS=b"\x02",
        P1=b"\x00",  # 0x00 - Return addres | 0x01 - Confirm befor ereturning
        P2=b"\x00",  # 0x00 - No chain code | 0x01 - With chain code
        Lc=len(DEFAULT_PATH_ENCODED).to_bytes(1, "big"),  # Payload byte length
        data=(len(DEFAULT_PATH_ENCODED) // 4).to_bytes(1, "big") + DEFAULT_PATH_ENCODED,
    )

    GET_ADDRESS_NO_CONFIRM = ISO7816Command(
        CLA=b"\xe0",
        INS=b"\x02",
        P1=b"\x00",  # 0x00 - Return addres | 0x01 - Confirm befor ereturning
        P2=b"\x00",  # 0x00 - No chain code | 0x01 - With chain code
        # Lc=len(DEFAULT_PATH_ENCODED).to_bytes(1, 'big'),  # Payload byte length
        # data=(len(DEFAULT_PATH_ENCODED) // 4).to_bytes(1, 'big') + DEFAULT_PATH_ENCODED,
    )

    SIGN_TX_FIRST_DATA = ISO7816Command(
        CLA=b"\xe0",
        INS=b"\x04",
        P1=b"\x00",  # 0x00 - First TX data block | 0x80 - Secondary data block
        P2=b"\x00",
    )

    SIGN_TX_SECONDARY_DATA = ISO7816Command(
        CLA=b"\xe0",
        INS=b"\x04",
        P1=b"\x80",  # 0x00 - First TX data block | 0x80 - Secondary data block
        P2=b"\x00",
    )

    SIGN_MESSAGE_FIRST_DATA = ISO7816Command(
        CLA=b"\xe0",
        INS=b"\x08",
        P1=b"\x00",  # 0x00 - First TX data block | 0x80 - Secondary data block
        P2=b"\x00",
    )

    SIGN_MESSAGE_SECONDARY_DATA = ISO7816Command(
        CLA=b"\xe0",
        INS=b"\x08",
        P1=b"\x80",  # 0x00 - First TX data block | 0x80 - Secondary data block
        P2=b"\x00",
    )

    SIGN_TYPED_DATA = ISO7816Command(
        CLA=b"\xe0",
        INS=b"\x0c",
        P1=b"\x00",
        P2=b"\x00",
    )

    @staticmethod
    def get(name: str) -> bytes:
        if not hasattr(LedgerCommands, name):
            raise ValueError("Command not available")
        cmd = getattr(LedgerCommands, name)
        return cmd.encode()

    @staticmethod
    def get_with_data(
        name: str, data: bytes, Lc: bytes = None, Le: bytes = None
    ) -> bytes:
        if not hasattr(LedgerCommands, name):
            raise ValueError("Command not available")
        cmd = getattr(LedgerCommands, name)
        cmd.set_data(data)
        if Lc is not None:
            cmd.Lc = Lc
        if Le is not None:
            cmd.Le = Le
        return cmd.encode()


def dongle_send(dongle: Dongle, command_string: str) -> bytes:
    """Send a command to the dongle"""
    hex_command = LedgerCommands.get(command_string)
    try:
        return dongle.exchange(hex_command)
    except CommException as err:
        raise LedgerError.transalate_comm_exception(err) from err


def dongle_send_data(
    dongle: Dongle, command_string: str, data: bytes, Lc: bytes = None, Le: bytes = None
) -> bytes:
    """Send a command with data to the dongle"""
    hex_command = LedgerCommands.get_with_data(command_string, data, Lc=Lc, Le=Le)
    try:
        return dongle.exchange(hex_command)
    except CommException as err:
        raise LedgerError.transalate_comm_exception(err) from err


def decode_response_version_from_config(confbytes: bytes) -> str:
    """Decode the string version from the bytearray response from Ledger device"""
    return "{}.{}.{}".format(
        confbytes[1],
        confbytes[2],
        confbytes[3],
    )


def decode_response_address(response):
    """Decode an address response from the dongle"""
    offset = 1 + response[0]
    address_encoded = response[offset + 1 : offset + 1 + response[offset]]
    return add_0x_prefix(address_encoded.decode("ascii"))


def is_usable_version(confbytes: bytes) -> bool:
    """Only tested on 1.2.4"""
    version = decode_response_version_from_config(confbytes)
    v_parts = version.split(".")
    ver = [int(s) for s in v_parts]
    # v9.9.9 is MockLedger
    return ver[0] == 9 or not any([ver[0] != 1, ver[1] < 2, ver[2] < 4])


def init_dongle(dongle: Dongle = None, debug: bool = False) -> Dongle:
    """Initialize the dongle and sanity check the connection"""
    global DONGLE_CACHE, DONGLE_CONFIG_CACHE

    # If not given, use cache if available
    if dongle is None and DONGLE_CACHE is None:
        try:
            DONGLE_CACHE = getDongle(debug)
        except CommException as err:
            raise LedgerError.transalate_comm_exception(err) from err

        # Sanity check the version
        if DONGLE_CONFIG_CACHE is None or dongle is not None:
            DONGLE_CONFIG_CACHE = dongle_send(DONGLE_CACHE, "GET_CONFIGURATION")

        if not is_usable_version(DONGLE_CONFIG_CACHE):
            raise NotImplementedError("Unsupported firmware version")

    return dongle or DONGLE_CACHE



import os
from typing import Any, Dict, Type

from eth_utils import remove_0x_prefix


def getenvint(key, default=0):
    """Get an int from en env var or use default"""
    try:
        return int(os.environ.get("MAX_ACCOUNTS_FETCH"))
    except TypeError:
        return default


# Chain ID to use if not given by user
DEFAULT_CHAIN_ID = 1

# Default accounts to fetch with get_accounts
DEFAULT_ACCOUNTS_FETCH = 3

# Number of accounts to fetch when looking up an account by address
MAX_ACCOUNTS_FETCH = getenvint("MAX_ACCOUNTS_FETCH", 5)

# Whether to use the legacy bip32 path derivation used by Ledger Chrome app
LEGACY_ACCOUNTS = os.getenv("LEDGER_LEGACY_ACCOUNTS") is not None

DEFAULT_PATH_STRING = "44'/60'/0'/0/0"
DEFAULT_PATH_ENCODED = (
    b"\x80\x00\x00,\x80\x00\x00<\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
)
if LEGACY_ACCOUNTS:
    DEFAULT_PATH_STRING = "44'/60'/0'/0"
    DEFAULT_PATH_ENCODED = b"\x80\x00\x00,\x80\x00\x00<\x80\x00\x00\x00\x00\x00\x00\x00"
DEFAULT_PATH = remove_0x_prefix(DEFAULT_PATH_ENCODED.hex())
VRS_RETURN_LENGTH = int(65).to_bytes(1, "big")

# Data size expected from Ledger
DATA_CHUNK_SIZE = 255

# Default "zero" values in EVM/Solidity
DEFAULTS: Dict[Type, Any] = {
    int: 0,
    bytes: b"",
}



from typing import List, Optional

from eth_utils import to_checksum_address

from ledgereth.comms import (
    Dongle,
    decode_response_address,
    dongle_send_data,
    init_dongle,
)
from ledgereth.constants import (
    DEFAULT_ACCOUNTS_FETCH,
    LEGACY_ACCOUNTS,
    MAX_ACCOUNTS_FETCH,
)
from ledgereth.objects import LedgerAccount
from ledgereth.utils import parse_bip32_path


def get_account_by_path(
    path_string: str, dongle: Optional[Dongle] = None
) -> LedgerAccount:
    """Return an account for a specific `BIP-44`_ derivation path

    :param path_string: (:code:`str`) - HID derivation path for the account to
        sign with.
    :param dongle: (:class:`ledgerblue.Dongle.Dongle`) -  The Dongle instance to
        use to communicate with the Ledger device
    :return: :class:`ledgereth.objects.LedgerAccount` instance for the given
        account

    .. _`BIP-44`: https://en.bitcoin.it/wiki/BIP_0044
    """
    dongle = init_dongle(dongle)
    path = parse_bip32_path(path_string)
    lc = len(path).to_bytes(1, "big")
    data = (len(path) // 4).to_bytes(1, "big") + path
    response = dongle_send_data(dongle, "GET_ADDRESS_NO_CONFIRM", data, Lc=lc)
    return LedgerAccount(path_string, decode_response_address(response))


def get_accounts(
    dongle: Optional[Dongle] = None, count: int = DEFAULT_ACCOUNTS_FETCH
) -> List[LedgerAccount]:
    """Return available accounts

    :param dongle: (:class:`ledgerblue.Dongle.Dongle`) -  The Dongle instance to
        use to communicate with the Ledger device
    :param count: (:code:`int`) - Amount of accounts to return
    :return: list of :class:`ledgereth.objects.LedgerAccount` instances found on
        the ledger
    """
    accounts = []
    dongle = init_dongle(dongle)

    for i in range(count):
        if LEGACY_ACCOUNTS:
            path_string = f"44'/60'/0'/{i}"
        else:
            path_string = f"44'/60'/{i}'/0/0"
        account = get_account_by_path(path_string, dongle)
        accounts.append(account)

    return accounts


def find_account(
    address: str, dongle: Optional[Dongle] = None, count: int = MAX_ACCOUNTS_FETCH
) -> Optional[LedgerAccount]:
    """Find an account by address

    :param address: (:class:`str`) - An address to look up
    :param dongle: (:class:`ledgerblue.Dongle.Dongle`) - The Dongle instance to
        use to communicate with the Ledger device
    :param count: (:code:`int`) - How deep in the derivation sequence to look
    :return: :class:`ledgereth.objects.LedgerAccount` instance if found on the
        Ledger
    """

    address = to_checksum_address(address)

    for account in get_accounts(dongle, count):
        if account.address == address:
            return account

    return None



from enum import IntEnum

from ledgerblue.commException import CommException


class LedgerErrorCodes(IntEnum):
    OK = 0x9000

    ##
    # User/input errors
    ##

    TX_TYPE_UNSUPPORTED = 0x6501
    INCORRECT_LENGTH = 0x6700
    # This may happen if first/secondary data is invalid or out of order or
    # unexpected p1/p2 values
    INVALID_TX_CHUNKS = 0x6B00
    CANCELED_BY_USER = 0x6982
    # Invalid data, transaction, or HD path, and a bit of a catch-all
    INVALID_DATA = 0x6A80
    APP_SLEEP = 0x6804
    APP_NOT_STARTED = 0x6D00
    DEVICE_LOCKED = 0x6B0C
    # "Plugins" allow resolution of function selectors and args for various
    # smart contracts.  Ref: https://blog.ledger.com/ethereum-plugins/
    PLUGIN_NOT_PRESENT = 0x6984

    ##
    # Internal errors
    ##

    INT_CONVERSION_ERROR = 0x6504
    OUTPUT_BUFFER_TOO_SMALL = 0x6502
    PLUGIN_ERROR = 0x6503
    # "Signsture/parser not initialized" in source
    DECLINED = 0x6985

    ##
    # Inferred errors found through discovery
    ##

    # ledgerblue default code
    UKNOWN = 0x6F00
    APP_NOT_FOUND = 0x6D02

    @classmethod
    def get_by_value(cls, val):
        try:
            return cls(val)
        except ValueError:
            return None


class LedgerError(Exception):
    message = "Unexpected Ledger error"

    def __init__(self, message=None):
        super().__init__(message or self.message)

    @classmethod
    def transalate_comm_exception(cls, exp: CommException):
        return ERROR_CODE_EXCEPTIONS.get(
            exp.sw,
            LedgerError(
                f"Unexpected error: {hex(exp.sw)} {LedgerErrorCodes.get_by_value(exp.sw) or 'UNKNOWN'}"
            ),
        )


class LedgerNotFound(LedgerError):
    message = "Unable to find Ledger device"


class LedgerLocked(LedgerError):
    message = "Ledger appears to be locked"


class LedgerAppNotOpened(LedgerError):
    message = "Expected Ledger Ethereum app not open"


class LedgerCancel(LedgerError):
    message = "Action cancelled by the user"


class LedgerInvalid(LedgerError):
    message = 'Invalid data sent to ledger or "blind signing" is not enabled'


ERROR_CODE_EXCEPTIONS = {
    LedgerErrorCodes.UKNOWN: LedgerNotFound,
    LedgerErrorCodes.DEVICE_LOCKED: LedgerLocked,
    LedgerErrorCodes.APP_SLEEP: LedgerAppNotOpened,
    LedgerErrorCodes.APP_NOT_STARTED: LedgerAppNotOpened,
    LedgerErrorCodes.APP_NOT_FOUND: LedgerAppNotOpened,
    LedgerErrorCodes.CANCELED_BY_USER: LedgerCancel,
    LedgerErrorCodes.DECLINED: LedgerCancel,
    LedgerErrorCodes.INVALID_DATA: LedgerInvalid,
}



import binascii
from typing import List, Optional, Tuple, Union

from eth_utils import decode_hex
from rlp import Serializable, encode

from ledgereth.comms import Dongle, dongle_send_data, init_dongle
from ledgereth.constants import DATA_CHUNK_SIZE, DEFAULT_CHAIN_ID, DEFAULT_PATH_STRING
from ledgereth.objects import (
    SerializableTransaction,
    SignedTransaction,
    SignedType1Transaction,
    SignedType2Transaction,
    Transaction,
    TransactionType,
    Type1Transaction,
    Type2Transaction,
)
from ledgereth.utils import (
    chunks,
    coerce_access_list,
    is_bip32_path,
    is_hex_string,
    parse_bip32_path,
)

Text = Union[str, bytes]


def sign_transaction(
    tx: Serializable,
    sender_path: str = DEFAULT_PATH_STRING,
    dongle: Optional[Dongle] = None,
) -> SignedTransaction:
    """Sign a :class:`rlp.Serializable` transaction object.  Compatible with
    ledgereth and web3.py transaction objects.

    :param tx: (:class:`rlp.Serializable`) - Serializable transaction object to
        sign
    :param sender_path: (:code:`str`) - HID derivation path for the account to
        sign with. Defaults to first account in the derivation path.
    :param dongle: (:class:`ledgerblue.Dongle.Dongle`) -  The Dongle instance to
        use to communicate with the Ledger device
    :return: :class:`ledgereth.objects.SignedTransaction` instance for
        transaction
    """
    given_dongle = dongle is not None
    dongle = init_dongle(dongle)
    retval = None

    if isinstance(tx, Transaction):
        encoded_tx = encode(tx, Transaction)
    elif isinstance(tx, Type1Transaction):
        encoded_tx = tx.transaction_type.to_byte() + encode(tx, Type1Transaction)
    elif isinstance(tx, Type2Transaction):
        encoded_tx = tx.transaction_type.to_byte() + encode(tx, Type2Transaction)
    else:
        raise ValueError(
            "Only Transaction and Type2Transaction objects are currently supported"
        )

    if not is_bip32_path(sender_path):
        raise ValueError("Invalid sender BIP32 path given to sign_transaction")

    path = parse_bip32_path(sender_path)
    payload = (len(path) // 4).to_bytes(1, "big") + path + encoded_tx

    chunk_count = 0
    for chunk in chunks(payload, DATA_CHUNK_SIZE):
        chunk_size = len(chunk)

        if chunk_count == 0:
            retval = dongle_send_data(
                dongle,
                "SIGN_TX_FIRST_DATA",
                chunk,
                Lc=chunk_size.to_bytes(1, "big"),
            )
        else:
            retval = dongle_send_data(
                dongle,
                "SIGN_TX_SECONDARY_DATA",
                chunk,
                Lc=chunk_size.to_bytes(1, "big"),
            )
        chunk_count += 1

    if retval is None or len(retval) < 64:
        raise Exception("Invalid response from Ledger")

    chain_id = tx.chain_id or DEFAULT_CHAIN_ID
    r = int(binascii.hexlify(retval[1:33]), 16)
    s = int(binascii.hexlify(retval[33:65]), 16)

    if tx.transaction_type < TransactionType.EIP_2930:
        if (chain_id * 2 + 35) + 1 > 255:
            ecc_parity = retval[0] - ((chain_id * 2 + 35) % 256)
            v = (chain_id * 2 + 35) + ecc_parity
        else:
            v = retval[0]

        signed = SignedTransaction(
            nonce=tx.nonce,
            gas_price=tx.gas_price,
            gas_limit=tx.gas_limit,
            destination=tx.destination,
            amount=tx.amount,
            data=tx.data,
            v=v,
            r=r,
            s=s,
        )
    else:
        y_parity = retval[0]

        if tx.transaction_type == TransactionType.EIP_2930:
            signed = SignedType1Transaction(
                chain_id=tx.chain_id,
                nonce=tx.nonce,
                gas_limit=tx.gas_limit,
                destination=tx.destination,
                amount=tx.amount,
                data=tx.data,
                gas_price=tx.gas_price,
                access_list=tx.access_list,
                y_parity=y_parity,
                sender_r=r,
                sender_s=s,
            )
        else:
            signed = SignedType2Transaction(
                chain_id=tx.chain_id,
                nonce=tx.nonce,
                gas_limit=tx.gas_limit,
                destination=tx.destination,
                amount=tx.amount,
                data=tx.data,
                max_priority_fee_per_gas=tx.max_priority_fee_per_gas,
                max_fee_per_gas=tx.max_fee_per_gas,
                access_list=tx.access_list,
                y_parity=y_parity,
                sender_r=r,
                sender_s=s,
            )

    # If this func inited the dongle, then close it, otherwise core dump
    if not given_dongle:
        dongle.close()

    return signed


def create_transaction(
    destination: Text,
    amount: int,
    gas: int,
    nonce: int,
    data: Text = b"",
    gas_price: int = 0,
    max_priority_fee_per_gas: int = 0,
    max_fee_per_gas: int = 0,
    chain_id: int = DEFAULT_CHAIN_ID,
    sender_path: str = DEFAULT_PATH_STRING,
    access_list: Optional[List[Tuple[Union[bytes, str], List[int]]]] = None,
    dongle: Optional[Dongle] = None,
) -> SignedTransaction:
    """Create and sign a transaction from given arguments.

    :param destination: (:code:`str|bytes`) - Destination address (AKA :code:`to`)
    :param amount: (:code:`int`) - Transaction value in wei
    :param gas: (:code:`int`) - Gas limit for the transaction
    :param nonce: (:code:`int`) - Nonce for the transaction
    :param data: (:code:`str|bytes`) - Transaction data (e.g. contract calldata)
    :param gas_price: (:code:`int`) - Gas price in wei to use for the
        transaction.  This is not compatible with :code:`max_fee_per_gas`.
    :param max_priority_fee_per_gas: (:code:`int`) - Priority fee per gas (in
        wei) to provide to the miner of the block.
    :param max_fee_per_gas: (:code:`int`) - Maximum fee in wei to pay for the
        transaction.  This is not compatible with :code:`gas_price`.
    :param chain_id: (:code:`int`) - Chain ID to limit the transaction to.
        Defaults to :code:`1`.
    :param sender_path: (:code:`str`) - `BIP-44`_ HD derivation path for the
        account to sign with. Defaults to first account in the default
        derivation path.
    :param access_list: (:code:`List[Tuple[bytes, List[int]]]`) -
        `EIP-2930`_ access list to use for the transaction.
    :param dongle: (:class:`ledgerblue.Dongle.Dongle`) -  The Dongle instance to
        use to communicate with the Ledger device
    :return: :class:`ledgereth.objects.SignedTransaction` instance for transaction

    .. _`BIP-44`: https://en.bitcoin.it/wiki/BIP_0044
    .. _`EIP-2930`: https://eips.ethereum.org/EIPS/eip-2930
    """
    given_dongle = dongle is not None
    dongle = init_dongle(dongle)

    if type(destination) == str and is_hex_string(destination):
        destination = decode_hex(destination)

    if not data:
        data = b""
    elif type(data) == str and is_hex_string(data):
        data = decode_hex(data)

    # be cool mypy
    assert type(destination) == bytes
    assert type(data) == bytes

    # EIP-1559 transactions should never have gas_price
    if gas_price and (max_priority_fee_per_gas or max_fee_per_gas):
        raise ValueError(
            "gas_price is incompatible with max_priority_fee_per_gas and max_fee_per_gas"
        )

    # we need a gas price for a valid transaction
    if not gas_price and not max_fee_per_gas:
        raise ValueError("gas_price or max_fee_per_gas must be provided")

    # Create a serializable tx object
    if max_fee_per_gas:
        tx = Type2Transaction(
            destination=destination,
            amount=amount,
            gas_limit=gas,
            data=data,
            nonce=nonce,
            chain_id=chain_id,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            max_fee_per_gas=max_fee_per_gas,
            access_list=coerce_access_list(access_list),
        )
    elif access_list is not None:
        tx = Type1Transaction(
            destination=destination,
            amount=amount,
            gas_limit=gas,
            data=data,
            nonce=nonce,
            chain_id=chain_id,
            gas_price=gas_price,
            access_list=coerce_access_list(access_list),
        )
    else:
        tx = Transaction(
            destination=destination,
            amount=amount,
            gas_limit=gas,
            gas_price=gas_price,
            data=data,
            nonce=nonce,
            chain_id=chain_id,
        )

    signed = sign_transaction(tx, sender_path, dongle=dongle)

    # If this func inited the dongle, then close it, otherwise core dump
    if not given_dongle:
        dongle.close()

    return signed


def decode_transaction(rawtx: bytes, signed: bool = False) -> SerializableTransaction:
    """Decode a raw transaction to a Serializable transaction object

    :param rawtx: (:code:`bytes`) - Raw transaction to decode
    :param signed: (:code:`bool`) - If the raw transaction is a signed
        transaction.
    :return: Decoded :class:`ledgereth.objects.SerializableTransaction` instance
        for transaction
    """
    tx_type = rawtx[0]

    tx = None

    if tx_type < 127:
        if tx_type == 1:
            if signed:
                return SignedType1Transaction.from_rawtx(rawtx)
            else:
                return Type1Transaction.from_rawtx(rawtx)
        elif tx_type == 2:
            if signed:
                return SignedType2Transaction.from_rawtx(rawtx)
            else:
                return Type2Transaction.from_rawtx(rawtx)
        else:
            raise NotImplementedError(
                f"Support for transaction type {tx_type} has not yet been implemented"
            )
    elif signed:
        return SignedTransaction.from_rawtx(rawtx)

    return Transaction.from_rawtx(rawtx)



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Any, Dict, List, Optional, Tuple

from eth_utils import encode_hex, to_checksum_address
from rlp import Serializable, decode, encode
from rlp.sedes import BigEndianInt, Binary, CountableList
from rlp.sedes import List as ListSedes
from rlp.sedes import big_endian_int, binary

from ledgereth.constants import DEFAULT_CHAIN_ID
from ledgereth.utils import (
    coerce_list_types,
    is_bip32_path,
    is_bytes,
    is_optional_bytes,
    parse_bip32_path,
)

address = Binary.fixed_length(20, allow_empty=False)
access_list_sede_type = CountableList(
    ListSedes(
        [
            address,
            CountableList(BigEndianInt(32)),
        ]
    ),
)
RPC_TX_PROP_TRANSLATION = {
    "gas_price": "gasPrice",
    "gas_limit": "gas",
    "amount": "value",
    "destination": "to",
    "max_priority_fee_per_gas": "maxPriorityFeePerGas",
    "max_fee_per_gas": "maxFeePerGas",
    "access_list": "accessList",
    "chain_id": "chainId",
}
RPC_TX_PROPS = [
    "chainId",
    "from",
    "to",
    "gas",
    "gasPrice",
    "value",
    "data",
    "nonce",
    "maxFeePerGas",
    "maxPriorityFeePerGas",
    "chainId",
    "accessList",
]


class TransactionType(IntEnum):
    """An Ethereum EIP-2718 transaction type"""

    #: Original and EIP-155
    LEGACY = 0
    #: Type-1 (Access Lists)
    EIP_2930 = 1
    #: Type-2 (Transaction fee change to max fee and priority fee)
    EIP_1559 = 2

    def to_byte(self):
        """Decode TransactionType to a single byte"""
        return self.value.to_bytes(1, "big")


class ISO7816Command:
    """A representation of an ISO-7816 APDU Command binary to be sent to the
    Ledger device."""

    def __init__(
        self,
        CLA: bytes,
        INS: bytes,
        P1: bytes,
        P2: bytes,
        Lc: bytes = None,
        Le: bytes = None,
        data: bytes = None,
    ):
        if not (
            is_bytes(CLA)
            and is_bytes(INS)
            and is_bytes(P1)
            and is_bytes(P2)
            and is_optional_bytes(Lc)
            and is_optional_bytes(Le)
            and is_optional_bytes(data)
        ):
            raise TypeError("Command parts must be type bytes")

        self.CLA = CLA
        self.INS = INS
        self.P1 = P1
        self.P2 = P2
        self.Lc = Lc or b"\x00"
        self.Le = Le
        self.data = data

    def set_data(self, data: bytes, Lc: bytes = None) -> None:
        """Set the command data and its length

        :param data: (:class:`bytes`) - The raw ``bytes`` data. This should not
            exceed the max chunk length of 255 (including command data)
        :param Lc: (:class:`bytes`) - The length of the data
        """
        self.data = data

        if len(self.data) > 255:
            # TODO: Warning?
            return

        if Lc is None:
            self.Lc = len(self.data).to_bytes(1, "big")
        else:
            self.Lc = Lc

    def encode(self) -> bytes:
        """Encode the command into ``bytes`` to be sent to the Ledger device.

        :return: Encoded ``bytes`` data
        """
        encoded = self.CLA + self.INS + self.P1 + self.P2

        if self.data is not None:
            if self.Lc is None:
                self.Lc = (len(self.data)).to_bytes(1, "big")
            encoded += self.Lc
            encoded += self.data
        else:
            encoded += self.Lc

        if self.Le is not None:
            encoded += self.Le

        return encoded

    def encode_hex(self) -> str:
        """Encode the command into hex bytes representation.

        :return: Encoded hex ``str``
        """
        return self.encode().hex()


class LedgerAccount:
    """A representation of an account derived from the private key on a Ledger
    device."""

    #: The HD path of the account
    path: str

    #: The HD path of the account
    path_encoded: bytes

    #: The account's address
    address: str

    def __init__(self, path, address):
        """Initialize an account.

        :param path: (``str``) Derivation path for the account
        :param address: (``str``) Address of the account
        """
        if not is_bip32_path(path):
            raise ValueError("Invalid BIP32 Ethereum path")

        self.path = path
        self.path_encoded = parse_bip32_path(path)
        self.address = to_checksum_address(address)

    def __repr__(self):
        return f"<ledgereth.objects.LedgerAccount {self.address}>"


class SerializableTransaction(Serializable):
    """An RLP Serializable transaction object"""

    @classmethod
    @abstractmethod
    def from_rawtx(cls, rawtx: bytes) -> SerializableTransaction:
        """Instantiates a SerializableTransaction given a raw encoded
        transaction

        :param rawtx: (:class:`bytes`) - The decoded raw transaction ``bytes``
            to encode into a :class`ledgereth.objects.SerializableTransaction`
        :return: Instantiated :class`ledgereth.objects.SerializableTransaction`
        """

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary representation of the transaction

        :return: Transaction dict
        """
        d = {}
        for name, _ in self.__class__._meta.fields:
            d[name] = getattr(self, name)
        return d

    def to_rpc_dict(self) -> Dict[str, Any]:
        """To a dict compatible with web3.py or JSON-RPC

        :return: Transaction dict
        """
        d: Dict[str, Any] = {}

        for name, _ in self.__class__._meta.fields:
            key = (
                RPC_TX_PROP_TRANSLATION[name]
                if name in RPC_TX_PROP_TRANSLATION
                else name
            )

            if key in RPC_TX_PROPS:
                # Need to format an access list differently for web3/RPC-like
                # objects.  It expects a list of objects
                if key == "accessList":
                    orig = getattr(self, name)
                    d[key] = []
                    for item in orig:
                        d[key].append(
                            {
                                "address": item[0],
                                "storageKeys": [
                                    int.from_bytes(slot, "big") for slot in item[1]
                                ],
                            }
                        )
                else:
                    d[key] = getattr(self, name)

        return d


class Transaction(SerializableTransaction):
    """Unsigned legacy or `EIP-155`_ transaction

    .. note:: A chain_id is set by default (``1``).  It is not required to be
        a valid legacy transaction, but without it your transaction is
        suceptible to replay attack.  If for some reason you absolutely do not want it in your
        tx, set it to ``None``.

    .. _`EIP-155`: https://eips.ethereum.org/EIPS/eip-155
    """

    fields = [
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas_limit", big_endian_int),
        ("destination", address),
        ("amount", big_endian_int),
        ("data", binary),
        ("chain_id", big_endian_int),
        # Expected nine elements as part of EIP-155 transactions
        ("dummy1", big_endian_int),
        ("dummy2", big_endian_int),
    ]

    #: The EIP-2718 transaction type
    transaction_type = TransactionType.LEGACY

    def __init__(
        self,
        nonce: int,
        gas_price: int,
        gas_limit: int,
        destination: bytes,
        amount: int,
        data: bytes,
        chain_id: int = DEFAULT_CHAIN_ID,
        dummy1: int = 0,
        dummy2: int = 0,
    ):
        """Initialize an unsigned transaction

        :param nonce: (``int``) Transaction nonce
        :param gas_price: (``int``) Gas price in wei
        :param gas_limit: (``int``) Gas limit
        :param destination: (``bytes``) Destination address
        :param amount: (``int``) Amount of Ether to send in wei
        :param data: (``bytes``) Transaction data
        :param chain_id: (``int``) Chain ID
        :param dummy1: (``int``) **DO NOT SET**
        :param dummy2: (``int``) **DO NOT SET**
        """
        super().__init__(
            nonce,
            gas_price,
            gas_limit,
            destination,
            amount,
            data,
            chain_id,
            dummy1,
            dummy2,
        )

    @classmethod
    def from_rawtx(cls, rawtx: bytes) -> Transaction:
        """Instantiate a Transaction object from a raw encoded transaction

        :param rawtx: (``bytes``) A raw transaction to instantiate with
        :returns: :class:`ledgereth.objects.Transaction`
        """
        if rawtx[0] < 127:
            raise ValueError("Transaction is not a legacy transaction")

        return Transaction(
            *coerce_list_types(
                [int, int, int, bytes, int, bytes, int, int, int], decode(rawtx)
            )
        )


class Type1Transaction(SerializableTransaction):
    """An unsigned Type 1 transaction.

    Encoded tx format spec:

    .. code::

        0x01 || rlp([chainId, nonce, gasPrice, gasLimit, destination, amount, data, accessList])
    """

    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas_limit", big_endian_int),
        ("destination", address),
        ("amount", big_endian_int),
        ("data", binary),
        ("access_list", access_list_sede_type),
    ]

    #: The EIP-2718 transaction type
    transaction_type = TransactionType.EIP_2930

    def __init__(
        self,
        chain_id: int,
        nonce: int,
        gas_price: int,
        gas_limit: int,
        destination: bytes,
        amount: int,
        data: bytes,
        access_list: Optional[List[Tuple[bytes, List[int]]]] = None,
    ):
        """Initialize an unsigned type 2 transaction

        :param chain_id: (``int``) Chain ID
        :param nonce: (``int``) Transaction nonce
        :param gas_price: (``int``) Gas price in wei
        :param gas_limit: (``int``) Gas limit
        :param destination: (``bytes``) Destination address
        :param amount: (``int``) Amount of Ether to send in wei
        :param data: (``bytes``) Transaction data
        :param access_list: (``Optional[List[Tuple[bytes, List[int]]]]``) EIP-2718 Access
            list
        """
        access_list = access_list or []
        super().__init__(
            chain_id,
            nonce,
            gas_price,
            gas_limit,
            destination,
            amount,
            data,
            access_list,
        )

    @classmethod
    def from_rawtx(cls, rawtx: bytes) -> Type1Transaction:
        """Instantiate a Type1Transaction object from a raw encoded transaction

        :param rawtx: (``bytes``) A raw transaction to instantiate with
        :returns: :class:`ledgereth.objects.Type1Transaction`
        """
        if rawtx[0] != cls.transaction_type:
            raise ValueError(
                f"Transaction is not a type {cls.transaction_type} transaction"
            )

        return Type1Transaction(
            *coerce_list_types(
                [int, int, int, int, bytes, int, bytes, None],
                decode(rawtx[1:]),
            )
        )


class Type2Transaction(SerializableTransaction):
    """An unsigned Type 2 transaction.

    Encoded TX format spec:

    .. code::

        0x02 || rlp([chain_id, nonce, max_priority_fee_per_gas, max_fee_per_gas, gas_limit, destination, amount, data, access_list])
    """

    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("max_priority_fee_per_gas", big_endian_int),
        ("max_fee_per_gas", big_endian_int),
        ("gas_limit", big_endian_int),
        ("destination", address),
        ("amount", big_endian_int),
        ("data", binary),
        ("access_list", access_list_sede_type),
    ]

    #: The EIP-2718 transaction type
    transaction_type = TransactionType.EIP_1559

    def __init__(
        self,
        chain_id: int,
        nonce: int,
        max_priority_fee_per_gas: int,
        max_fee_per_gas: int,
        gas_limit: int,
        destination: bytes,
        amount: int,
        data: bytes,
        access_list: Optional[List[Tuple[bytes, List[int]]]] = None,
    ):
        """Initialize an unsigned type 2 transaction

        :param chain_id: (``int``) Chain ID
        :param nonce: (``int``) Transaction nonce
        :param max_priority_fee_per_gas: (``int``) Priority fee per gas (in
            wei) to provide to the miner of the block.
        :param max_fee_per_gas: (``int``) Maximum fee in wei to pay for the
            transaction.  This is not compatible with :code:`gas_price`.
        :param gas_limit: (``int``) Gas limit
        :param destination: (``bytes``) Destination address
        :param amount: (``int``) Amount of Ether to send in wei
        :param data: (``bytes``) Transaction data
        :param access_list: (``List[Tuple[bytes, List[int]]]``) EIP-2718 Access
            list
        """
        access_list = access_list or []
        super().__init__(
            chain_id,
            nonce,
            max_priority_fee_per_gas,
            max_fee_per_gas,
            gas_limit,
            destination,
            amount,
            data,
            access_list,
        )

    @classmethod
    def from_rawtx(cls, rawtx: bytes) -> Type2Transaction:
        """Instantiate a Type2Transaction object from a raw encoded transaction

        :param rawtx: (``bytes``) A raw transaction to instantiate with
        :returns: :class:`ledgereth.objects.Type2Transaction`
        """
        if rawtx[0] != cls.transaction_type:
            raise ValueError(
                f"Transaction is not a type {cls.transaction_type} transaction"
            )

        return Type2Transaction(
            *coerce_list_types(
                [int, int, int, int, int, bytes, int, bytes, None],
                decode(rawtx[1:]),
            )
        )


class SignedTransaction(SerializableTransaction):
    """Signed legacy or EIP-155 transaction"""

    fields = [
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas_limit", big_endian_int),
        ("destination", address),
        ("amount", big_endian_int),
        ("data", binary),
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    ]

    #: The EIP-2718 transaction type
    transaction_type = TransactionType.LEGACY

    def __init__(
        self,
        nonce: int,
        gas_price: int,
        gas_limit: int,
        destination: bytes,
        amount: int,
        data: bytes,
        v: int,
        r: int,
        s: int,
    ):
        """Initialize an unsigned transaction

        :param nonce: (``int``) Transaction nonce
        :param gas_price: (``int``) Gas price in wei
        :param gas_limit: (``int``) Gas limit
        :param destination: (``bytes``) Destination address
        :param amount: (``int``) Amount of Ether to send in wei
        :param data: (``bytes``) Transaction data
        :param v: (``int``) Signature v value
        :param r: (``int``) Signature r value
        :param s: (``int``) Signature s value
        """
        super().__init__(
            nonce, gas_price, gas_limit, destination, amount, data, v, r, s
        )

    @classmethod
    def from_rawtx(cls, rawtx: bytes) -> SignedTransaction:
        """Instantiate a SignedTransaction object from a raw encoded transaction

        :param rawtx: (``bytes``) A raw signed transaction to instantiate with
        :returns: :class:`ledgereth.objects.SignedTransaction`
        """
        if rawtx[0] < 127:
            raise ValueError("Transaction is not a legacy transaction")

        return SignedTransaction(
            *coerce_list_types([int, int, int, int, bytes, int, bytes], decode(rawtx))
        )

    def raw_transaction(self):
        """Return an encoded raw signed transaction

        Encoded signed TX format spec:

        .. code::

            rlp([nonce, gasPrice, gasLimit, destination, amount, data, signatureV, signatureR, signatureS])

        :returns: Encoded raw signed transaction bytes
        """
        return encode_hex(encode(self, SignedTransaction))

    # Match the API of the web3.py Transaction object
    #: Encoded raw signed transaction
    rawTransaction = property(raw_transaction)


class SignedType1Transaction(SerializableTransaction):
    """A signed Type 1 transaction."""

    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas_limit", big_endian_int),
        ("destination", address),
        ("amount", big_endian_int),
        ("data", binary),
        ("access_list", access_list_sede_type),
        ("y_parity", big_endian_int),
        ("sender_r", big_endian_int),
        ("sender_s", big_endian_int),
    ]

    #: The EIP-2718 transaction type
    transaction_type = TransactionType.EIP_2930

    def __init__(
        self,
        chain_id: int,
        nonce: int,
        gas_price: int,
        gas_limit: int,
        destination: bytes,
        amount: int,
        data: bytes,
        access_list: List[Tuple[bytes, List[int]]],
        y_parity: int,
        sender_r: int,
        sender_s: int,
    ):
        """Initialize a signed type 1 transaction

        :param chain_id: (``int``) Chain ID
        :param nonce: (``int``) Transaction nonce
        :param gas_price: (``int``) Gas price in wei
        :param gas_limit: (``int``) Gas limit
        :param destination: (``bytes``) Destination address
        :param amount: (``int``) Amount of Ether to send in wei
        :param data: (``bytes``) Transaction data
        :param access_list: (``List[Tuple[bytes, List[int]]]``) EIP-2718 Access
            list
        :param y_parity: (``int``) Parity byte for the signature
        :param sender_r: (``int``) Signature r value
        :param sender_s: (``int``) Signature s value
        """
        super().__init__(
            chain_id,
            nonce,
            gas_price,
            gas_limit,
            destination,
            amount,
            data,
            access_list,
            y_parity,
            sender_r,
            sender_s,
        )

    @classmethod
    def from_rawtx(cls, rawtx: bytes) -> SignedType1Transaction:
        """Instantiate a SignedType1Transaction object from a raw encoded
        transaction

        :param rawtx: (``bytes``) A raw signed transaction to instantiate with
        :returns: :class:`ledgereth.objects.SignedType1Transaction`
        """
        if rawtx[0] != cls.transaction_type:
            raise ValueError(
                f"Transaction is not a type {cls.transaction_type} transaction"
            )

        return SignedType1Transaction(
            *coerce_list_types(
                [int, int, int, int, bytes, int, bytes, None, int, int, int],
                decode(rawtx[1:]),
            )
        )

    def raw_transaction(self):
        """Return an encoded raw signed transaction

        Encoded signed TX format spec:

        .. code::

            0x01 || rlp([chainId, nonce, gasPrice, gasLimit, destination, amount, data, accessList, signatureYParity, signatureR, signatureS])

        :returns: Encoded raw signed transaction bytes
        """
        return encode_hex(b"\x01" + encode(self, SignedType1Transaction))

    # Match the API of the web3.py Transaction object
    #: Encoded raw signed transaction
    rawTransaction = property(raw_transaction)


class SignedType2Transaction(SerializableTransaction):
    """A signed Type 2 transaction."""

    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("max_priority_fee_per_gas", big_endian_int),
        ("max_fee_per_gas", big_endian_int),
        ("gas_limit", big_endian_int),
        ("destination", address),
        ("amount", big_endian_int),
        ("data", binary),
        ("access_list", access_list_sede_type),
        ("y_parity", big_endian_int),
        ("sender_r", big_endian_int),
        ("sender_s", big_endian_int),
    ]

    #: The EIP-2718 transaction type
    transaction_type = TransactionType.EIP_1559

    def __init__(
        self,
        chain_id: int,
        nonce: int,
        max_priority_fee_per_gas: int,
        max_fee_per_gas: int,
        gas_limit: int,
        destination: bytes,
        amount: int,
        data: bytes,
        access_list: List[Tuple[bytes, List[int]]],
        y_parity: int,
        sender_r: int,
        sender_s: int,
    ):
        """Initialize a signed type 2 transaction

        :param chain_id: (``int``) Chain ID
        :param nonce: (``int``) Transaction nonce
        :param max_priority_fee_per_gas: (``int``) Priority fee per gas (in
            wei) to provide to the miner of the block.
        :param max_fee_per_gas: (``int``) Maximum fee in wei to pay for the
            transaction.  This is not compatible with :code:`gas_price`.
        :param gas_limit: (``int``) Gas limit
        :param destination: (``bytes``) Destination address
        :param amount: (``int``) Amount of Ether to send in wei
        :param data: (``bytes``) Transaction data
        :param access_list: (``List[Tuple[bytes, List[int]]]``) EIP-2718 Access
            list
        :param y_parity: (``int``) Parity byte for the signature
        :param sender_r: (``int``) Signature r value
        :param sender_s: (``int``) Signature s value
        """
        super().__init__(
            chain_id,
            nonce,
            max_priority_fee_per_gas,
            max_fee_per_gas,
            gas_limit,
            destination,
            amount,
            data,
            access_list,
            y_parity,
            sender_r,
            sender_s,
        )

    @classmethod
    def from_rawtx(cls, rawtx: bytes) -> SignedType2Transaction:
        """Instantiate a SignedType2Transaction object from a raw encoded
        transaction

        :param rawtx: (``bytes``) A raw signed transaction to instantiate with
        :returns: :class:`ledgereth.objects.SignedType2Transaction`
        """
        if rawtx[0] != cls.transaction_type:
            raise ValueError(
                f"Transaction is not a type {cls.transaction_type} transaction"
            )

        return SignedType2Transaction(
            *coerce_list_types(
                [
                    int,
                    int,
                    int,
                    int,
                    int,
                    bytes,
                    int,
                    bytes,
                    None,
                    int,
                    int,
                    int,
                ],
                decode(rawtx[1:]),
            )
        )

    def raw_transaction(self):
        """Return an encoded raw signed transaction

        Encoded signed TX format spec:

        ..code::

            0x02 || rlp([chain_id, nonce, max_priority_fee_per_gas, max_fee_per_gas, gas_limit, destination, amount, data, access_list, signature_y_parity, signature_r, signature_s])

        :returns: Encoded raw signed transaction bytes
        """
        return encode_hex(b"\x02" + encode(self, SignedType2Transaction))

    # Match the API of the web3.py Transaction object
    #: Encoded raw signed transaction
    rawTransaction = property(raw_transaction)


class Signed(ABC):
    #: Signature v
    v: int
    #: Signature r
    r: int
    #: Signature s
    s: int

    def __init__(self, v, r, s):
        self.v = v
        self.r = r
        self.s = s

    @property
    def signature(self):
        """Encoded signature

        :returns: Signature ``bytes``
        """
        if not self.v or not self.r or not self.s:
            raise ValueError("Missing v, r, or s")

        return encode_hex(
            self.r.to_bytes(32, "big")
            + self.s.to_bytes(32, "big")
            + self.v.to_bytes(1, "big")
        )


class SignedMessage(Signed):
    """Signed EIP-191 message"""

    message: bytes

    def __init__(self, message, v, r, s):
        """Initialize a singed message

        :param message: (``bytes``) Message that was signed
        :param v: (``int``) Signature v value
        :param r: (``int``) Signature r value
        :param s: (``int``) Signature s value
        """
        self.message = message
        super().__init__(v, r, s)


class SignedTypedMessage(Signed):
    """Signed EIP-812 typed data"""

    domain_hash: bytes
    message_hash: bytes

    def __init__(self, domain_hash, message_hash, v, r, s):
        """Initialize a singed message

        :param domain_hash: (``bytes``) Domain hash that was signed
        :param message_hash: (``bytes``) Message hash that was signed
        :param v: (``int``) Signature v value
        :param r: (``int``) Signature r value
        :param s: (``int``) Signature s value
        """
        self.domain_hash = domain_hash
        self.message_hash = message_hash
        super().__init__(v, r, s)



from importlib.machinery import SourceFileLoader
from pathlib import Path
from subprocess import check_call

from setuptools import find_packages, setup
from setuptools.command.develop import develop

pwd = Path(__file__).parent

# Get the long description from the README file
with pwd.joinpath("README.md").open(encoding="utf-8") as f:
    long_description = f.read()


def requirements_to_list(filename):
    return [
        dep
        for dep in pwd.joinpath(filename).open(encoding="utf-8").read().split("\n")
        if (dep and not dep.startswith("#"))
    ]


class LintCommand(develop):
    """Run linting"""

    def run(self):
        try:
            check_call("black .".split())
            check_call("isort --profile black .".split())
            check_call("autoflake -ir .".split())
        except CalledProcessError as err:
            if "non-zero" in str(err):
                print("linting failed with warnings", file=sys.stderr)
                sys.exit(1)


# Allows us to import the file without executing imports in module __init__
meta = SourceFileLoader("meta", str(pwd.joinpath("ledgereth/_meta.py"))).load_module()

setup(
    name="ledgereth",
    version=meta.version,
    description="Library to interface with ledger-app-eth on Ledger hardware wallets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mikeshultz/ledger-eth-lib",
    author=meta.author,
    author_email=meta.email,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="solidity ethereum development",
    packages=find_packages(exclude=["docs", "tests", "scripts", "build"]),
    install_requires=requirements_to_list("requirements.txt"),
    extras_require={
        "dev": requirements_to_list("requirements.dev.txt"),
        "docs": requirements_to_list("requirements.docs.txt"),
    },
    package_data={
        "": [
            "README.md",
            "LICENSE",
        ],
    },
    cmdclass={
        "lint": LintCommand,
    },
)



import re
import struct
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from eth_utils import decode_hex

from ledgereth.constants import DEFAULTS

# 44'/60'/0'/0/x
BIP32_ETH_PATTERN = r"^44'/60'/[0-9]+'/[0-9]+/[0-9]+$"
BIP32_LEGACY_LEDGER_PATTERN = r"^44'/60'/[0-9]+'/[0-9]+$"

COERCERS: Dict[Type, Callable] = {int: lambda v: int.from_bytes(v, "big")}


def is_bytes(v: Any) -> bool:
    return isinstance(v, bytes)


def is_optional_bytes(v: Optional[Any]) -> bool:
    return v is None or is_bytes(v)


def is_hex_string(v: Any) -> bool:
    return isinstance(v, str) and v.startswith("0x")


def is_bip32_path(path: str) -> bool:
    """Detect if a string a bip32 path that can be given to a Ledger device"""
    return (
        re.match(BIP32_ETH_PATTERN, path) is not None
        or re.match(BIP32_LEGACY_LEDGER_PATTERN, path) is not None
    )


def chunks(it: bytes, chunk_size: int) -> Generator[bytes, None, None]:
    """Iterate bytes(it) into chunks of chunk_size"""

    if not isinstance(it, bytes):
        raise TypeError("iterable argument must be type bytes")

    it_size = len(it)

    if it_size <= chunk_size:
        yield it
    else:
        chunk_count, remainder = divmod(it_size, chunk_size)
        for i in range(0, chunk_count * chunk_size, chunk_size):
            yield it[i : i + chunk_size]
        final_offset = chunk_count * chunk_size
        yield it[final_offset : final_offset + remainder]


def parse_bip32_path(path: str) -> bytes:
    if not path:
        return b""

    result = b""
    elements = path.split("/")

    for pathElement in elements:
        element = pathElement.split("'")

        # Wihout tick (') == 1
        if len(element) == 1:
            # "public" BIP-44 derivation
            result = result + struct.pack(">I", int(element[0]))
        else:
            # "private" BIP-44 derivation
            result = result + struct.pack(">I", 0x80000000 | int(element[0]))

    return result


def decode_bip32_path(path: bytes) -> str:
    """Decode a BIP-32/44 path from bytes"""
    parts = []

    for i in range(0, len(path) // 4):
        idx = i * 4
        chunk = path[idx : idx + 4]
        result = struct.unpack(">I", chunk)

        if result[0] < 256:
            # "public" BIP-44 derivation
            parts.append(f"{result[0]}")
        else:
            # "private" BIP-44 derivation
            part = result[0] - (0x80000000 & result[0])
            parts.append(f"{part}'")

    return "/".join(parts)


def decode_access_list(
    access_list: Collection[Tuple[bytes, Collection[bytes]]]
) -> List[Tuple[bytes, Tuple[int, ...]]]:
    """Decode an access list into friendly Python types"""
    work_list = []

    if not access_list or len(access_list) < 1:
        return []

    for item in access_list:
        work_slot_list = []

        for slot in item[1]:
            work_slot_list.append(int.from_bytes(slot, "big"))

        work_list.append((item[0], tuple(work_slot_list)))

    return work_list


def decode_web3_access_list(
    access_list: Collection[Dict[str, Union[str, Collection[str]]]]
) -> List[Tuple[bytes, Tuple[int, ...]]]:
    work_list = []

    if not access_list or len(access_list) < 1:
        return []

    for idx, item in enumerate(access_list):
        if "address" not in item:
            raise ValueError(f"Access list item at position {idx} missing address")

        if "storageKeys" not in item:
            raise ValueError(f"Access list item at position {idx} missing storageKeys")

        assert type(item["address"]) == str

        work_list.append(
            (
                decode_hex(item["address"]),
                tuple(
                    map(
                        lambda v: int.from_bytes(decode_hex(v), "big"),
                        tuple(item.get("storageKeys", [])),
                    )
                ),
            )
        )

    return work_list


def coerce_access_list(access_list):
    """Validate and type coerce an access list from Python friendly values to
    values for RLP encoding"""
    if access_list is None:
        return []

    if type(access_list) != list:
        raise ValueError("Expected access_list to be a list")

    for i, rule in enumerate(access_list):
        if type(rule) not in (list, tuple):
            raise ValueError("Expected access_list rules to be a list or tuple")

        if type(rule) == tuple:
            access_list[i] = list(rule)

        target, slots = rule

        if is_hex_string(target):
            access_list[i][0] = decode_hex(target)
        elif type(target) != bytes:
            raise ValueError(
                f"Unexpected type ({type(target)}) for access_list address at index {i}"
            )

        for j, slot in enumerate(slots):
            if is_hex_string(slot):
                access_list[i][1][j] = int(slot, 16)
            elif type(slot) != int:
                raise ValueError(
                    f"Unexpected type ({type(slot)}) for access_list slot at index {j}"
                )

    return access_list


def coerce_list_types(types: List[Type], to_coerce: List[Any]) -> List[Any]:
    """Coerce types of a list to given types in order"""

    for i, v in enumerate(to_coerce):
        # SKIP!
        if types[i] is None:
            continue

        # Some things don't transalate, like b'' being 0
        if not v:
            to_coerce[i] = DEFAULTS[types[i]]
        else:
            if types[i] in COERCERS:
                to_coerce[i] = COERCERS[types[i]](v)
            else:
                to_coerce[i] = types[i](v)

    return to_coerce
